#--
#                 - Mellanox Confidential and Proprietary -
# 
# Copyright (C) Dec 2020, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
# 
# Except as specifically permitted herein, no portion of the information,
# including but not limited to object code and source code, may be reproduced,
# modified, distributed, republished or otherwise exploited in any form or by
# any means for any purpose without the prior written permission of Mellanox
# Technologies Ltd. Use of software subject to the terms and conditions
# detailed in the file "LICENSE.txt".
#--

import sys
if sys.version_info.major < 3:
    import commands as sub
else:
    import subprocess as sub
import os
import glob
import re
from mtcr import MstDevice
import math
import regaccess

DEV_TYPE_REG_ADDR = int("0xf0014", 16)
GBOX_DIR          = '/dev/mst/gbox'
MTUSB_DEV_PATH    = '/dev/mst/'
GBOX_MNGR_SUFIX   = "_gbox_mngr"
DEV_PATH          = "/dev/"
LC_PREFIX         = "/dev/lc"
GEARBOX_PREFIX    = "gearbox0"
MANAGER_PREFIX    = "fpga"
GBOX_STR          = "_gbox53108"
SYSTEM_BOARD_TYPE = "/sys/devices/virtual/dmi/id/board_name"


GB_DEV_ID          = int("0x00000252", 16)
GB_MNGR_DEV_ID     = int("0x00000253", 16)
SPECTRUM_3_DEV_ID  = int("0x00000250", 16)
DEVICE_ID_MASK     = 0x0000FFFF

GB_MNGR_SLAVE_ADDR = int("0x33", 16)

DEV_SW_PTRN = re.compile(r'.+(mt5\d{4}_pciconf\d)')

QUERY_GEARBOX   = 2
QUERY_LINECARD  = 1
GB_MNGR_INDEX   = 4
MAX_SLOTS       = 8

def check_if_device_gbox(device_name):
    dev = MstDevice(device_name)

    could_read = True
    try :
        output = dev.read4(DEV_TYPE_REG_ADDR)
        output = output & DEVICE_ID_MASK
    except :
        could_read = False
    dev.close()

    if could_read and output == GB_DEV_ID:
        return True
    return False
    # commnad_str = "mcra {} 0xf0014".format(device_name)


def check_if_device_spectrum3(mst_device):
    could_read = True
    try :
        output = mst_device.read4(DEV_TYPE_REG_ADDR)
        output = output & DEVICE_ID_MASK
    except :
        could_read = False
    mst_device.close()

    if could_read and output == SPECTRUM_3_DEV_ID:
        return True
    return False


def check_if_device_spectrum4(mst_device):
    could_read = True
    try :
        output = mst_device.read4(DEV_TYPE_REG_ADDR)
    except :
        could_read = False
    mst_device.close()

    if could_read and output == SPECTRUM_4_DEV_ID:
        return True
    return False


def check_if_device_gbox_mngr(device_name):
    dev = MstDevice(device_name)
    could_read = True
    output = None
    try :
        dev.msetI2cSlave(GB_MNGR_SLAVE_ADDR)
        output = dev.read4(DEV_TYPE_REG_ADDR)
        output = output & DEVICE_ID_MASK
    except :
        could_read = False
    dev.close()

    if could_read and output == GB_MNGR_DEV_ID:
        return True
    return False
    # commnad_str = "mcra -s 0x33 {}_gbox 0xf0014".format(device_name)


def create_dir (dir_name):
    if not os.path.isdir(dir_name):
        try:
            os.mkdir(dir_name)
        except OSError:
            print ("-E- Creation of the directory %s failed" % dir_name)


def create_gbox_dir():
    create_dir (GBOX_DIR)


def delete_gbox_empty_dir():
    if os.path.isdir(GBOX_DIR):
        try:
            os.rmdir(GBOX_DIR)
        except OSError:
            print ("-E- Deletion of the directory %s failed" % GBOX_DIR)

def is_gbox_dit_empty():
    return (len(os.listdir(GBOX_DIR)) == 0)
       
def create_file(file_name, parent_device):
    if not os.path.exists(file_name) :
        #os.mknod(file_name)
        parent_file = MTUSB_DEV_PATH + parent_device
        if os.path.islink(parent_file):
            link_name = os.readlink(parent_file)
            os.symlink(link_name, file_name)


def create_files_for_gbox(parent_device, is_mngr):
    if is_mngr:
        file_name = GBOX_DIR + "/" + parent_device  + GBOX_STR + '_ln0_' + 'mngr'
        # check first manager file do not exist
        if not os.path.exists(file_name):
            create_file(file_name, parent_device)
            return 1
    else:
        # check first file with same parent was not created yet
        if len(glob.glob(GBOX_DIR + "/" + parent_device + "*")) == 0:
            for i in range(0,4):
                file_name = GBOX_DIR + "/" + parent_device  + GBOX_STR + '_ln0_' + str(i)
                create_file(file_name, parent_device)
            return 4
    return 0


def get_mtusb_devices():
    gbox_list = []
    gbox_mngr_list = []

    files_list = glob.glob(MTUSB_DEV_PATH + "*")
    for dev in files_list:
        # if mtusb device - is link
        if os.path.islink(dev):
            # check if gbox or mngr
            if check_if_device_gbox(dev) or check_if_device_gbox_mngr(dev):
                dev_name = dev.replace(MTUSB_DEV_PATH, '')
                gbox_list.append(dev_name)
                gbox_mngr_list.append(dev_name)

    return gbox_list, gbox_mngr_list


def create_mtusb_gb_devices():
    gb_devs_created = 0
    gbm_devs_created = 0

    gbox_list, gbox_mngr_list = get_mtusb_devices()

    if (len(gbox_list) + len(gbox_mngr_list)) > 0:
        create_gbox_dir()
        # create gbox 0-3 for each mtusb
        for dev in gbox_list:
            gb_devs_created += create_files_for_gbox(dev, False)
        # # create mngr for mtusb
        for dev in gbox_mngr_list:
            gbm_devs_created += create_files_for_gbox(dev, True)

    return gb_devs_created, gbm_devs_created


def is_bufallo():
    cmd = "cat " + SYSTEM_BOARD_TYPE

    rc, output = sub.getstatusoutput(cmd)
    if not rc:
        return (output == "VMOD0011")
    return False


def create_i2c_gb_devices():
    #if not buffalo - return 0 devices
    if not is_bufallo():
        return 0, 0

    # /dev/lc1/gearbox01
    created_devices = 0
    created_mngr_devices = 0
    dirs = glob.glob(DEV_PATH + "*")
    detected_lcs = []

    create_gbox_dir()

    # get lcs dettected
    for dir_name in dirs:
        if LC_PREFIX in dir_name:
            lc_index = dir_name.replace(LC_PREFIX, '')
            gb_files = glob.glob(dir_name + "/*") 

            for gb_file in gb_files:
                add_mngr = 0
                add_gb = 0
                if GEARBOX_PREFIX in gb_file:
                    file_name = os.path.basename(gb_file)
                    gb_index = file_name.replace(GEARBOX_PREFIX, '')

                    new_file = GBOX_DIR + "/i2c" + GBOX_STR + "_ln" + str(lc_index) + "_" + str(gb_index)
                    add_gb = 1
                if MANAGER_PREFIX in gb_file:
                    # if mng 
                    new_file = GBOX_DIR + "/i2c" + GBOX_STR + "_ln" + str(lc_index) + "_mngr"
                    add_mngr = 1
                if (add_gb + add_mngr) > 0 and not os.path.exists(new_file) :
                    os.symlink(gb_file, new_file)
                    created_devices += add_gb
                    created_mngr_devices += add_mngr

    return created_devices, created_mngr_devices


def get_devices_on_slot(reg_access, slot_index):
    devices_on_slot = []
    request_message_sequence = 0
    while True:
        try:
            mddq_get = reg_access.sendMddq(QUERY_GEARBOX, request_message_sequence, slot_index)
            if mddq_get[0]["data_valid"]:
                if mddq_get[1]["device_type"] == 0: # 0 for GB
                    devices_on_slot.append(mddq_get[1]["device_index"])
        except regaccess.RegAccException:
            break
        request_message_sequence += 1
    return devices_on_slot


def query_switch_gbs(mst_str):
    reg_access = regaccess.RegAccess(pci_device=mst_str)
    # query_type: (1: slot information, 2: devices on slot))
    # request_message_sequence: (slots or devices indicator, if set, we have a device)
    # slot_index: line card index
    # query_index: not related to GBs
    slot_index=1
    slots={}
    switch_code = DEV_SW_PTRN.match(mst_str).group(1)
    # query slot information
    try:
        while slot_index <= MAX_SLOTS:
            mddq_get = reg_access.sendMddq(QUERY_LINECARD, 0, slot_index)
            if (mddq_get[1]["lc_ready"] == 1) or mddq_get[1]["active"]:
                slots.update({slot_index:get_devices_on_slot(reg_access, slot_index)})
            slot_index += 1
    except regaccess.RegAccException:
        pass

    num_of_gboxs = 0
    num_of_mgboxs = 0
    for lc in slots.keys():
        for gbox in slots[lc]:
            sfx = str(gbox)
            if gbox == GB_MNGR_INDEX:
                sfx = "mngr"
                num_of_mgboxs += 1
            else:
                num_of_gboxs += 1
            file_name = GBOX_DIR + "/" + "switch_" + switch_code + GBOX_STR + "_ln" + str(lc) + "_" + sfx
            # if file do not exist - create it as link
            if file_name and not os.path.exists(file_name) :
                os.symlink(mst_str, file_name)

    return num_of_gboxs, num_of_mgboxs


def create_switch_gb_devices():
    num_of_switch_gb_devs = 0
    for dev in glob.glob("/dev/mst/gbox/*"):
        if 'switch_' in dev:
            num_of_switch_gb_devs += 1

    switches_list = []
    for dev in glob.glob("/dev/mst/*"):
        if DEV_SW_PTRN.match(dev):
            switches_list.append(dev)

    create_gbox_dir()

    num_of_gboxs = 0
    num_of_mgboxs = 0
    for switch in switches_list:
        mst_dev = MstDevice(switch)
        if not (check_if_device_spectrum3(mst_dev) or check_if_device_spectrum4(mst_dev)):
            mst_dev.close()
            continue
        mst_dev.close()
        rc = query_switch_gbs(switch)
        num_of_gboxs += rc[0]
        num_of_mgboxs += rc[1]

    num_of_new_gbs = num_of_gboxs - num_of_switch_gb_devs
    
    return num_of_new_gbs, num_of_mgboxs


if __name__ == '__main__':

    gb_mtusb_devs_count, gbm_mtusb_devs_count = create_mtusb_gb_devices()

    gb_i2c_devs_count, gbm_i2c_devs_count = create_i2c_gb_devices()

    gb_switch_devs_count, gbm_switch_devs_count = create_switch_gb_devices()

    gb_devs_count = gb_mtusb_devs_count + gb_i2c_devs_count + gb_switch_devs_count
    gbm_devs_count = gbm_mtusb_devs_count + gbm_i2c_devs_count + gbm_switch_devs_count

    if is_gbox_dit_empty():
        delete_gbox_empty_dir()
    if gb_devs_count > 0:
        print ("-I- {0} Gearbox devices added".format(gb_devs_count))
    else: 
        print ("-I- No new Gearbox devices found")
    if gbm_devs_count > 0:
        print ("-I- {0} Gearbox Mannager devices added".format(gbm_devs_count))
